from django.shortcuts import render, get_object_or_404
from django.views.generic import DetailView, ListView, CreateView
from django.contrib.auth.decorators import login_required
from django.urls import reverse
from django.http import HttpResponseRedirect, HttpResponse, Http404
from django.core.exceptions import ValidationError, PermissionDenied
from django.contrib.auth.models import Group
from guardian.decorators import permission_required
from guardian.mixins import PermissionRequiredMixin, PermissionListMixin
from guardian.shortcuts import assign_perm, get_users_with_perms, get_objects_for_user

from . import models
from . import forms

class MaskRelatedWrapper:
    def __init__(self, orig_obj, user, related_permissions):
        self.orig_obj = orig_obj
        self.user = user
        self.related_permissions = related_permissions

    def _has_mask(self, field):
        return field in self.related_permissions

    def _get_perm(self, field):
        return self.related_permissions[field]

    def __getattr__(self, key):
        orig_attr = getattr(self.orig_obj, key)
        if self._has_mask(key):
            perm = self._get_perm(key)
            return get_objects_for_user(self.user, perm, orig_attr)
        return orig_attr

    def __getattribute__(self, name):
        if name == "__class__":
            return self.orig_obj.__class__
        return object.__getattribute__(self, name)

class RelatedPermissionRequiredMixin:
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        if not hasattr(self, "related_permission_required"):
            raise TypeError("RelatedPermissionRequiredMixin needs "
                            "related_permission_required field")

    def get_object(self, *args, **kwds):
        obj = super().get_object(*args, **kwds)
        masked_obj = MaskRelatedWrapper(obj,
                                        self.request.user,
                                        self.related_permission_required)
        return masked_obj

class CaListView(PermissionListMixin, ListView):
    model = models.CertificationAuthority
    permission_required = ['view_certificationauthority']


class CaDetailView(RelatedPermissionRequiredMixin, PermissionRequiredMixin, DetailView):
    model = models.CertificationAuthority
    permission_required = ['view_certificationauthority']

    related_permission_required = {
        "issued_certificates": "view_certificate",
        "requests": "view_certificatesigningrequest",
    }

class CsrListView(PermissionListMixin, ListView):
    model = models.CertificateSigningRequest
    permission_required = ['view_certificatesigningrequest']

class CertListView(PermissionListMixin, ListView):
    model = models.Certificate
    permission_required = ['view_certificate']

class CertDetailView(PermissionRequiredMixin, DetailView):
    model = models.Certificate
    permission_required = ['view_certificate']

@permission_required('view_certificate', (models.Certificate, "pk", "pk"))
def cert_download(request, pk):
    cert = get_object_or_404(models.Certificate, pk=pk)
    response = HttpResponse(cert.pem, content_type="application/x-pem-file")
    response['Content-Disposition'] = 'inline; filename=cert.pem'
    return response

class CertPickupView(DetailView):
    slug_field = 'pickup_code'
    slug_url_kwarg = 'pickup_code'
    template_name = "owlca/certificate_pickup.html"
    model = models.CertificateSigningRequest

def cert_pickup_download(request, pickup_code, cert_pk):
    csr = get_object_or_404(models.CertificateSigningRequest,
                            pickup_code=pickup_code)
    try:
        cert = csr.certificates.get(pk=cert_pk)
    except (ValueError, models.Certificate.DoesNotExist) as e:
        raise Http404()

    response = HttpResponse(cert.pem, content_type="application/x-pem-file")
    response['Content-Disposition'] = 'inline; filename=cert.pem'
    return response

@permission_required('owlca.add_certificationauthority')
def ca_create(request):
    if request.method == 'POST':
        form = forms.CaCreateForm(request.POST)
        if form.is_valid():
            password = form.cleaned_data['password'].encode()

            ca = models.CertificationAuthority()
            ca.title = form.cleaned_data['title']
            ca.comment = form.cleaned_data['comment']
            ca.country = form.cleaned_data['country']
            ca.state_or_province = form.cleaned_data['state_or_province']
            ca.locality = form.cleaned_data['locality']
            ca.organization = form.cleaned_data['organization']
            ca.common_name = form.cleaned_data['common_name']

            ca.generate_key(
                password,
                length=form.cleaned_data['length']
            )
            ca.save()

        
            assign_perm('view_certificationauthority', request.user, ca)
            assign_perm('manage_certificationauthority', request.user, ca)
            assign_perm('change_certificationauthority', request.user, ca)
            assign_perm('request_certificate', request.user, ca)
            if form.cleaned_data['public']:
                any_user = Group.objects.get(name="any-user")
                assign_perm('view_certificationauthority',
                            any_user, ca)
                assign_perm('request_certificate',
                            any_user, ca)

            csr = ca.create_self_sign_request(password)
            assign_perm('view_certificatesigningrequest', request.user, csr)
            assign_perm('change_certificatesigningrequest', request.user, csr)

            cert = ca.approve(csr, password, days=30 * 365)
            ca.cert = cert
            assign_perm('view_certificate', request.user, cert)
            assign_perm('change_certificate', request.user, cert)
            if form.cleaned_data['public']:
                any_user = Group.objects.get(name="any-user")
                assign_perm('view_certificate',
                            any_user, cert)

            ca.save()

            return HttpResponseRedirect(reverse('ca-detail', args=[ca.pk]))

    # if a GET (or any other method) we'll create a blank form
    else:
        form = forms.CaCreateForm()

    return render(request,
                  'owlca/certificationauthority_create.html',
                  {'form': form})

@permission_required('owlca.change_certificationauthority')
def ca_edit(request, pk):
    ca = get_object_or_404(models.CertificationAuthority, pk=pk)
    if request.method == 'POST':
        form = forms.CaEditForm(request.POST)
        if form.is_valid():
            old_password = form.cleaned_data['old_password'].encode()
            new_password = form.cleaned_data['new_password'].encode()
            repeat = form.cleaned_data['repeat'].encode()

            try:
                ca.change_password(old_password, new_password)
                ca.save()

                return HttpResponseRedirect(reverse('ca-detail', args=[ca.pk]))
            except ValueError:
                form.add_error("old_password", "Decryption of CA key failed")


    # if a GET (or any other method) we'll create a blank form
    else:
        form = forms.CaEditForm()

    return render(request,
                  'owlca/certificationauthority_edit.html',
                  {'form': form, 'certificationauthority': ca})

def inherit_perm(origin, origin_perm, target, target_perms):
    """Grant target_perm@target to all users who have origin_perm@origin"""
    authorized_users = get_users_with_perms(
        origin,
        with_group_users=False,
        only_with_perms_in=[origin_perm]
    )

    for target_perm in target_perms:
        assign_perm(target_perm, authorized_users, target)

@permission_required('request_certificate',
                     (models.CertificationAuthority, "pk", "pk"))
def csr_create(request, pk):
    ca = get_object_or_404(models.CertificationAuthority, pk=pk)
    if request.method == 'POST':
        form = forms.CsrCreateForm(request.POST, request.FILES)
        form_helper = forms.CsrCreateHelperForm(request.POST)

        if form.is_valid() or form_helper.is_valid():
            if form_helper.is_valid():
                csr_pem = form_helper.cleaned_data["csr_pem"]
            else:
                csr_pem = request.FILES['csr_pem'].read().decode()

            csr = models.CertificateSigningRequest()
            csr.pem = csr_pem
            csr.ca = ca
            csr.save()

            inherit_perm(ca, "manage_certificationauthority",
                         csr, ["view_certificatesigningrequest",
                               "change_certificatesigningrequest"])

            return HttpResponseRedirect(reverse('cert-pickup',
                                        args=[csr.pickup_code]))
    else:
        form = forms.CsrCreateForm()

    return render(request,
                  'owlca/certificatesignrequest_create.html',
                  {'form': form, 'certificationauthority': ca})

@permission_required("view_certificatesigningrequest",
                     (models.CertificateSigningRequest, "pk", "pk"))
def csr_detail(request, pk):
    csr = get_object_or_404(models.CertificateSigningRequest, pk=pk)

    if request.method == 'POST':
        if not request.user.has_perm("manage_certificationauthority", csr.ca):
            raise PermissionDenied()
                
        form = forms.CsrResolveForm(request.POST)
        if form.is_valid():
            if "approve" in form.data:
                password = form.cleaned_data["password"].encode()
                days = form.cleaned_data["days"]

                ca = csr.ca
                try:
                    cert = ca.approve(csr, password, days=days)

                    inherit_perm(ca, "manage_certificationauthority",
                                 cert, ["view_certificate",
                                        "change_certificate"])

                    return HttpResponseRedirect(reverse('cert-detail',
                                                args=[cert.pk]))
                except ValueError as e:
                    form.errors["password"] = [ValidationError("Signing failed")]
            else:
                csr.verification = models.VERIFICATION_FAILED
                csr.save()
                return HttpResponseRedirect(reverse('csr-detail',
                                            args=[csr.pk]))

    else:
        form = forms.CsrResolveForm()

    return render(request,
                  'owlca/certificatesigningrequest_detail.html',
                  {'form': form, 'certificatesigningrequest': csr})

